# -*- coding: utf-8 -*-
"""
Celery tasks related models
"""
import os
import hashlib
from datetime import (
    datetime,
    timedelta,
)
from pyramid.security import (
    Allow,
)
from sqlalchemy import (
    Integer,
    Column,
    ForeignKey,
    String,
    DateTime,
    Text,
)
from endi_base.models.types import (
    JsonEncodedList,
    PersistentACLMixin,
    ACLType,
    MutableList,
)
from endi_base.models.base import (
    DBBASE,
    default_table_args,
)
from sqlalchemy.orm import (
    relationship,
)


TIMEOUT = timedelta(seconds=10000)


class Job(DBBASE, PersistentACLMixin):
    """
    Base job model, used to communicate a job's status between the main pyramid
    app and celery asynchronous tasks
    """
    __tablename__ = 'job'
    __table_args__ = default_table_args
    __mapper_args__ = {
        'polymorphic_on': 'type_',
        'polymorphic_identity': 'job',
    }
    label = u"Tâche générique"
    id = Column(
        Integer,
        primary_key=True,
    )
    jobid = Column(
        String(255),
        nullable=True,
    )
    status = Column(
        String(20),
        default='planned',
    )
    created_at = Column(
        DateTime(),
        default=datetime.now,
    )
    updated_at = Column(
        DateTime(),
        default=datetime.now,
        onupdate=datetime.now
    )
    type_ = Column(
        'type_',
        String(30),
        nullable=False,
    )
    _acl = Column(
        MutableList.as_mutable(ACLType),
    )

    def todict(self):
        return dict(
            label=self.label,
            jobid=self.jobid,
            status=self.status,
            created_at=self.created_at.strftime("%d/%m/%Y à %H:%M"),
            updated_at=self.updated_at.strftime("%d/%m/%Y à %H:%M"),
        )

    def __json__(self, request):
        return self.todict()

    def set_owner(self, login):
        """
        Set the job owner's acls
        """
        self._acl = [(Allow, login, ('add', 'edit', 'view',))]

    def timeout(self):
        """
        Check if this element should be timeouted
        """
        result = True
        if self.status in ('planned', 'running'):
            if datetime.now() - self.updated_at > TIMEOUT:
                self.status = u"failed"
                if hasattr(self, 'error_messages'):
                    self.error_messages = [
                        u"Cette tâche a été automatiquement annulée car elle "
                        u"n'a pas pu être traitée. Veulliez contacter un "
                        u"administrateur en lui fournissant l'identifiant de "
                        u"tâche suivant : {0}".format(self.id)
                    ]
                result = False
        return result


class CsvImportJob(Job):
    """
    Store csv importation job status
    """
    __tablename__ = 'csv_import_job'
    __table_args__ = default_table_args
    __mapper_args__ = {'polymorphic_identity': 'csv_import'}
    id = Column(Integer, ForeignKey('job.id'), primary_key=True)
    messages = Column(JsonEncodedList, default=None)
    error_messages = Column(JsonEncodedList, default=None)
    in_error_csv = Column(Text(), default=None)
    unhandled_datas_csv = Column(Text(), default=None)
    label = u"Import de données"

    def is_not_void_str(self, value):
        """
        Return True if the string contains datas
        """
        return not (value is None or len(value) == 0)

    def todict(self):
        res = Job.todict(self)
        res['label'] = self.label
        res['messages'] = self.messages
        res['error_messages'] = self.error_messages
        res['has_errors'] = self.is_not_void_str(self.in_error_csv)
        res['has_unhandled_datas'] = self.is_not_void_str(
            self.unhandled_datas_csv
        )
        return res


class MailingJob(Job):
    """
    Store mailing job status
    """
    __tablename__ = 'mailing_job'
    __table_args__ = default_table_args
    __mapper_args__ = {'polymorphic_identity': 'mailing'}
    id = Column(Integer, ForeignKey('job.id'), primary_key=True)
    messages = Column(JsonEncodedList, default=None)
    error_messages = Column(JsonEncodedList, default=None)
    label = u"Envoi de mail"

    def todict(self):
        res = Job.todict(self)
        res['label'] = self.label
        res['messages'] = self.messages
        res['error_messages'] = self.error_messages
        return res


class FileGenerationJob(Job):
    __tablename__ = 'file_generation_job'
    __table_args__ = default_table_args
    __mapper_args__ = {'polymorphic_identity': 'file_generation'}
    id = Column(Integer, ForeignKey('job.id'), primary_key=True)
    filename = Column(Text)
    error_messages = Column(JsonEncodedList, default=None)
    label = u"Génération de fichier"

    def todict(self):
        res = Job.todict(self)
        res['filename'] = self.filename
        res['error_messages'] = self.error_messages
        return res


class MailHistory(DBBASE):
    """
    Stores the history of mail sent by our application to any company
    """
    __tablename__ = 'mail_history'
    __table_args__ = default_table_args
    id = Column(Integer, primary_key=True)
    send_at = Column(
        DateTime(),
        default=datetime.now,
    )

    filepath = Column(String(255))
    md5sum = Column(String(100))
    company_id = Column(ForeignKey('company.id'), nullable=True)
    company = relationship(
        "Company",
    )

    @property
    def filename(self):
        return os.path.basename(self.filepath)


def store_sent_mail(filepath, filedatas, company_id):
    """
    Stores a sent email in the history

    :param filename: The path to the sent file
    :param filedatas: The file datas
    :param int company_id: the id of a company instance
    """
    mail_history = MailHistory(
        filepath=filepath,
        md5sum=hashlib.md5(filedatas).hexdigest(),
        company_id=company_id
    )
    return mail_history


def check_if_mail_sent(filedatas, company_id):
    """
    Check if the given file has already been sent
    :param str filedatas: The content of a file
    :param int company_id: The id of a company
    """
    query = MailHistory.query()
    query = query.filter(MailHistory.company_id == company_id)
    md5sum = hashlib.md5(filedatas).hexdigest()
    query = query.filter(MailHistory.md5sum == md5sum)
    return query.first() is not None



# -*- coding: utf-8 -*-
"""
Accounting operations parsing

Parses :

    csv files
    slk files

- Collect the administration mail address
- Inspect the pool for waiting files
- Find the parser and the filetype
- Parse
    - Collect file informations
    - Ensure we have csv datas
    - Insert lines in database

"""
import datetime
import os
import re
import transaction
import hashlib

from sylk_parser import SylkParser

from pyramid_celery import celery_app
from endi_base.mail import send_mail
from endi_base.utils import csv_tools, date as date_utils
from endi_base.models.base import DBSESSION
from endi_base.utils.math import convert_to_float
from endi.models.company import Company
from endi.models.accounting.operations import (
    AccountingOperationUpload,
    AccountingOperation,
)
from endi_celery.conf import (
    get_setting,
    get_recipients_addresses,
)
from endi_celery.tasks import utils
from endi_celery.tasks.accounting_measure_compute import (
    get_measure_compilers
)


logger = utils.get_logger(__name__)
FILENAME_ERROR = (
    u"Le fichier ne respecte pas la nomenclature de nom "
    u"supportée par enDI ex : \n"
    u" Pour les états de trésorerie : "
    u"'2017_12_01_balance_analytique.slk'.\n"
    u"Pour les comptes de résultat : \n"
    u"2017_12_resultat.slk",
)


def _get_base_path():
    """
    Retreive the base working path as configured in the ini file
    """
    return get_setting('endi.parsing_pool_parent', mandatory=True)


def _get_path(directory):
    """
    Return the abs path for the given directory

    :param str directory: The directory name pool/error/processed
    :rtype: str
    """
    return os.path.join(_get_base_path(), directory)


def _get_file_path_from_pool(pool_path):
    """
    Handle file remaining in the pool

    :param str pool_path: The pool path to look into
    :returns: The name of the first file we find in the rep
    :rtype: str
    """
    result = None
    if os.path.isdir(pool_path):
        files = os.listdir(pool_path)
        for file_ in files:
            path = os.path.join(pool_path, file_)
            if os.path.isfile(path):
                result = path
                break
    return result


def _get_md5sum(file_path,  blocksize=65536):
    """
    Return a md5 sum of the given file_path informations
    """
    hash = hashlib.md5()
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(blocksize), b""):
            hash.update(block)
    return hash.hexdigest()


def _mv_file(file_path, queue='processed'):
    """
    Move the file to the processed directory
    """
    if os.path.isfile(file_path):
        new_dir = _get_path(queue)
        new_file_path = os.path.join(new_dir, os.path.basename(file_path))
        os.system('mv "{0}" "{1}"'.format(file_path, new_file_path))
        logger.info("The file {0} has been moved to the {1} directory".format(
            file_path,
            new_dir,
        ))
        return new_file_path
    else:
        raise Exception(u"File is missing {}".format(file_path))


def _clean_old_operations(old_ids):
    """
    Clean old AccountingOperation entries
    :param list old_ids: The ids of items to remove
    """
    logger.info(u"  + Cleaning {0} old operations".format(len(old_ids)))
    op = AccountingOperation.__table__.delete().where(
        AccountingOperation.id.in_(old_ids)
    )
    op.execute()


class KnownError(Exception):
    pass


class AccountingDataParser(object):
    """
    Base Accounting datas parser
    """
    _supported_extensions = ('csv', 'slk')
    quotechar = '"'
    delimiter = ','
    encoding = 'utf-8'

    # To be filled in subclasses
    _filename_re = None
    filetype = None

    def __init__(self, file_path, force=False):
        self.file_path = file_path
        self.force = force
        self._file_datas = {}
        self._collect_main_file_infos()
        self.basename = self._file_datas['basename']
        self.company_id_cache = {}
        if hasattr(self, '_collect_specific_file_infos'):
            self._collect_specific_file_infos()

    def _load_company_id_cache(self):
        query = DBSESSION().query(
            Company.id, Company.code_compta
        ).filter(
            Company.code_compta != None  # noqa: E711
        )
        for id_, code in query:
            self.company_id_cache[code] = id_

    @classmethod
    def match(cls, filename):
        """
        Return True if the current parser should be able to handle that type of
        file
        """
        re_match = cls._filename_re.match(filename)
        return re_match is not None

    def _collect_main_file_infos(self):
        """
        Collect main informations about the current file

        basename extension (without the leading dot) md5sum
        :returns: A dict containing collected datas
        :rtype: dict
        """
        basename, extension = os.path.splitext(
            os.path.basename(self.file_path)
        )
        if extension:
            extension = extension[1:]
        if extension not in self._supported_extensions:
            raise KnownError(u"L'extension du fichier est inconnue")

        self._file_datas['basename'] = basename
        self._file_datas['extension'] = extension
        self._file_datas['md5sum'] = _get_md5sum(self.file_path)
        self._file_datas['filetype'] = self.filetype
        return self._file_datas

    def _stream_slk(self):
        """
        Stream the datas coming from a slk file
        :returns: An iterator for the sheet's lines
        """
        return SylkParser(self.file_path, use_unicode=True)

    def _stream_csv(self):
        """
        Stream csv datas
        :returns: An iterator of sheet lines as lists
        """
        with open(self.file_path) as fbuf:
            for line in csv_tools.UnicodeReader(
                fbuf, quotechar=self.quotechar, delimiter=self.delimiter,
                encoding=self.encoding,
            ):
                yield line

    def _stream_datas(self):
        """
        Stream the datas coming from a slk file
        :returns: An iterator for the sheets line
        """
        extension = self._file_datas['extension']

        func = getattr(self, "_stream_%s" % extension, None)
        if func is not None:
            for line in func():
                yield line

    def _find_company_id(self, analytical_account):
        """
        Find a company object starting from its analytical_account

        :param str analytical_account: The account
        :returns: The company's id
        """
        return self.company_id_cache.get(analytical_account)

    def _build_operation_upload_object(self):
        """
        Build an AccountingOperationUpload instance with current file datas

        :returns: An AccountingOperationUpload for the current parsed file
        """
        return AccountingOperationUpload(
            filename=os.path.basename(self.file_path),
            date=self._file_datas['date'],
            md5sum=self._file_datas['md5sum'],
            filetype=self._file_datas['filetype'],
        )

    def _build_operation(self, line_datas):
        """
        Build an AccountingOperation with the given line datas

        :param list line_datas: List of datas found in the line
        :returns: An instance of AccountingOperation
        """
        raise NotImplemented(
            u"_build_operation should be implemented in subclasses"
        )

    def _build_operations(self):
        """
        Build AccountingOperation based on the current file's datas

        :returns: A 2-uple with operations and number of missed associations
        :rtype: tuple (list, int)
        """
        operations = []
        missed_associations = 0
        for line in self._stream_datas():
            operation = self._build_operation(line)
            if operation is not None:
                operations.append(operation)
                if operation.company_id is None:
                    missed_associations += 1
        return operations, missed_associations

    def _already_loaded(self):
        """
        Check if the current file has already been loaded

        :rtype: bool
        """
        query = DBSESSION().query(AccountingOperationUpload.id)
        query = query.filter_by(md5sum=self._file_datas['md5sum'])
        return query.count() > 0

    def _get_existing_operation_ids(self):
        """
        Return ids of the operations already stored in database
        """
        return [
            entry[0] for entry in DBSESSION().query(AccountingOperation.id)
        ]

    def _get_num_val(self, line, index):
        """
        Retreieve the numeric value found at index in the line_datas list

        :param list line: List of datas coming from parsed file
        :param int index: The index in which we should retrieve the datas
        """
        result = 0
        if len(line) > index:
            result = line[index].strip() or 0
        return result

    def process_file(self):
        """
        Process file parsing

        :returns: The new AccountingOperationUpload instance and the number of
        missed associations (lines where we didn't found any matching company)
        :rtype: 2-uple
        """
        if self.force or not self._already_loaded():
            self._load_company_id_cache()
            old_ids = self._get_existing_operation_ids()
            upload_object = self._build_operation_upload_object()
            operations, missed_associations = self._build_operations()
            logger.info(
                u"Storing {0} new operations in database".format(
                    len(operations)
                )
            )
            logger.info(
                "  + {0} operations were not associated to an existing "
                "company".format(missed_associations)
            )
            if operations:
                upload_object.operations = operations
            else:
                old_ids = []

            DBSESSION().add(upload_object)
            DBSESSION().flush()

            return upload_object.id, missed_associations, old_ids

        else:
            logger.error(u"File {0} already loaded".format(self.file_path))
            raise KnownError(
                u"Ce fichier a déjà été traité : {0}".format(self.file_path)
            )


class GeneralLedgerParser(AccountingDataParser):
    _filename_re = re.compile(
        'general_ledger_'
        '(?P<year>[0-9]+)'
        '_(?P<month>[^_]+)'
        '_(?P<doctype>[^_]+)',
        re.IGNORECASE
    )
    filetype = 'general_ledger'

    def _collect_specific_file_infos(self):
        """
        Return specific file info
        """
        re_match = self._filename_re.match(self.basename)
        if re_match is None:
            raise KnownError(FILENAME_ERROR)

        logger.info(u"It's an General Ledger file")
        self._file_datas.update(re_match.groupdict())
        self._file_datas['date'] = datetime.date(
            int(self._file_datas['year']), int(self._file_datas['month']), 1
        )
        return self._file_datas

    def _build_operation(self, line_datas):
        """
        Build an AccountingOperation object

        :param list line_datas: List of datas found in the line
        :returns: An instance of AccountingOperation
        """
        result = None
        if len(line_datas) >= 6:
            if line_datas[0].strip() not in (
                u"Compte analytique de l'entrepreneur",
                u"Numéro analytique",
            ):
                analytical_account = line_datas[0].strip()
                general_account = line_datas[1].strip()
                date = date_utils.str_to_date(line_datas[2].strip())
                if not date:
                    logger.error(
                        u"This line has incorrect datas : %s" % line_datas
                    )
                    return None
                label = line_datas[5].strip()
                debit = self._get_num_val(line_datas, index=6)
                credit = self._get_num_val(line_datas, index=7)
                balance = 0
                company_id = self._find_company_id(analytical_account)
                result = AccountingOperation(
                    analytical_account=analytical_account,
                    general_account=general_account,
                    date=date,
                    label=label,
                    debit=convert_to_float(debit),
                    credit=convert_to_float(credit),
                    balance=convert_to_float(balance),
                    company_id=company_id,
                )
        else:
            if line_datas:
                logger.error(
                    u"This line is missing informations : %s" % line_datas
                )
        return result


class AnalyticalBalanceParser(AccountingDataParser):
    _filename_re = re.compile(
        'analytical_balance_'
        '(?P<year>[0-9]+)'
        '_(?P<month>[0-9]+)'
        '_(?P<day>[0-9]+)'
        '_(?P<doctype>[^_]+)',
        re.IGNORECASE
    )
    filetype = 'analytical_balance'

    def _collect_specific_file_infos(self):
        """
        Collect key datas from the file path
        """
        re_match = self._filename_re.match(self.basename)
        if re_match is None:
            raise KnownError(FILENAME_ERROR)

        logger.info(u"It's an Analytical Balance file")
        self._file_datas.update(re_match.groupdict())
        self._file_datas['date'] = datetime.date(
            int(self._file_datas['year']),
            int(self._file_datas['month']),
            int(self._file_datas['day']),
        )

        return self._file_datas

    def _build_operation(self, line_datas):
        """
        Build an AccountingOperation object

        :param list line_datas: List of datas found in the line
        :returns: An instance of AccountingOperation
        """
        result = None
        if len(line_datas) >= 5:
            if line_datas[0] not in (
                u"Compte analytique de l'entrepreneur",
                u"Numéro analytique",
            ):
                analytical_account = line_datas[0].strip()
                general_account = line_datas[2].strip()
                label = line_datas[3].strip()
                debit = self._get_num_val(line_datas, index=4)
                credit = self._get_num_val(line_datas, index=5)
                balance = self._get_num_val(line_datas, index=6)
                company_id = self._find_company_id(analytical_account)
                result = AccountingOperation(
                    analytical_account=analytical_account,
                    general_account=general_account,
                    label=label,
                    debit=convert_to_float(debit),
                    credit=convert_to_float(credit),
                    balance=convert_to_float(balance),
                    company_id=company_id,
                )
        else:
            if line_datas:
                logger.error(
                    u"This line is missing informations : %s" % line_datas
                )
        return result


MAIL_ERROR_SUBJECT = u"[ERREUR] enDI : traitement de votre document \
{filename}"

MAIL_ERROR_BODY = u"""Une erreur est survenue lors du traitement du
fichier {filename}:

    {error}
"""
MAIL_UNKNOWN_ERROR_BODY = u"""Une erreur inconnue est survenue lors du
traitement du fichier {filename}:

    {error}

Veuillez contacter votre administrateur
"""
MAIL_SUCCESS_SUBJECT = u"""enDI : traitement de votre document {0}"""
MAIL_SUCCESS_BODY = u"""Le fichier {0} a été traité avec succès.
Écritures générées : {1}
Écritures n'ayant pas pu être associées à une entreprise existante dans
enDI : {2}

Les indicateurs ont été générés depuis ces écritures.
"""


def send_error(request, mail_addresses, filename, err):
    """
    Send an error email to mail_addresses
    """
    if mail_addresses:
        try:
            message = MAIL_ERROR_BODY.format(
                error=err.message,
                filename=filename
            )
            subject = MAIL_ERROR_SUBJECT.format(filename=filename)
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except:
            logger.exception("send_success error")


def send_unknown_error(request, mail_addresses, filename, err):
    if mail_addresses:
        try:
            subject = MAIL_ERROR_SUBJECT.format(filename=filename)
            print(err.message)
            message = MAIL_UNKNOWN_ERROR_BODY.format(
                error=err.message,
                filename=filename
            )
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except:
            logger.exception("send_success error")


def send_success(request, mail_addresses, filename, new_entries, missing):
    if mail_addresses:
        try:
            subject = MAIL_SUCCESS_SUBJECT.format(filename)
            message = MAIL_SUCCESS_BODY.format(
                filename,
                new_entries,
                missing,
            )
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except:
            logger.exception("send_success error")


def _get_parser_factory(filename):
    """
    Find out which type of file we handle and return the associated parser

    :param str filename: The filename of the file to parse
    """
    result = None
    # ATtention l'ordre est important (le deuxième matche les fichiers qui
    # doivent être traités par le premier, mais pas l'inverse)
    for parser in (AnalyticalBalanceParser, GeneralLedgerParser):
        if parser.match(filename):
            result = parser
            break
    return result


def _move_file_to_processing(waiting_file):
    """
    MOve the waiting file to the processing queue

    :param str waiting_file: The full path to the file to process
    """
    logger.info(u" + Moving the file to the processing directory")
    file_to_parse = _mv_file(waiting_file, "processing")
    return file_to_parse


def _is_processing():
    """
    Check if there is already a parsing process running

    :rtype: bool
    """
    path = _get_path('processing')

    for filename in os.listdir(path):
        filepath = os.path.join(path, filename)
        if os.path.isfile(filepath):
            return True
    return False


@celery_app.task(bind=True)
def handle_pool_task(self, force=False):
    """
    Parse the files present in the configured file pool
    """
    pool_path = _get_path('pool')
    waiting_file = _get_file_path_from_pool(pool_path)

    if waiting_file is None:
        return

    already_processing = _is_processing()

    if already_processing:
        logger.info(u"A parsing is already processing")
        return

    else:
        file_to_parse = _move_file_to_processing(waiting_file)
        logger.info(u"Parsing an accounting file : %s" % file_to_parse)

        mail_addresses = get_recipients_addresses(self.request)
        filename = os.path.basename(file_to_parse)
        parser_factory = _get_parser_factory(filename)
        if parser_factory is None:
            err = Exception(
                u"Type de fichier inconnu, le nom du fichier ne "
                u"respecte pas les nomenclatures de nommage des "
                u"fichiers de trésorerie"
            )
            send_error(self.request, mail_addresses, filename, err)
            logger.error(u"Incorrect file type : %s" % filename)
            _mv_file(file_to_parse, "error")
            return False
        try:
            parser = parser_factory(file_to_parse, force)

            transaction.begin()
            logger.info(u"  + Storing accounting operations in database")
            upload_object_id, missed_associations, old_ids = \
                parser.process_file()
            logger.debug(u"  + File was processed")
            transaction.commit()
        except KnownError as err:
            transaction.abort()
            logger.exception(u"KnownError : %s" % err.message)
            logger.exception(u"* FAILED : transaction has been rollbacked")
            if mail_addresses:
                send_error(self.request, mail_addresses, filename, err)
                logger.error(
                    u"An error mail has been sent to {0}".format(mail_addresses)
                )
            _mv_file(file_to_parse, 'error')
            logger.error(u"File has been moved to error directory")
            return False

        except Exception as err:
            transaction.abort()
            logger.exception(u"Unkown Error")
            logger.exception(u"* FAILED : transaction has been rollbacked")
            if mail_addresses:
                send_unknown_error(self.request, mail_addresses, filename, err)
                logger.error(
                    u"An error mail has been sent to {0}".format(mail_addresses)
                )
            _mv_file(file_to_parse, 'error')
            logger.error(u"File has been moved to error directory")
            return False

        else:
            logger.info(u"Accounting operations where successfully stored")
            _mv_file(file_to_parse)
            logger.info(u"File has been moved to processed directory")

        if old_ids:
            transaction.begin()
            logger.info("  + Cleaning %s old operations" % len(old_ids))
            try:
                _clean_old_operations(old_ids)
                transaction.commit()
            except:
                transaction.abort()
                logger.exception(u"Error while cleaning old operations")
            else:
                logger.info(u" * Old datas cleaned successfully")

        transaction.begin()
        logger.info(u" + Compiling the measures")
        # On rafraichit l'objet car nous sommes dans une nouvelle transaction
        # on évite ainsi le problème de "is not bound to a Session"
        upload_object = AccountingOperationUpload.get(upload_object_id)
        num_operations = len(upload_object.operations)

        logger.debug(" + Retrieved the upload object %s" % upload_object.date)
        logger.debug(" + %s operations" % num_operations)

        measure_compiler_factories = get_measure_compilers(parser.filetype)

        for measure_compiler_factory in measure_compiler_factories:
            try:
                measure_compiler = measure_compiler_factory(
                    upload_object, upload_object.operations
                )
                measure_compiler.process_datas()
            except KnownError as err:
                transaction.abort()
                logger.exception(u"KnownError : %s" % err.message)
                if mail_addresses:
                    send_error(self.request, mail_addresses, filename, err)
                    logger.error(
                        u"An error mail has been sent to {0}".format(
                            mail_addresses
                        )
                    )
                logger.exception(u"* FAILED : transaction has been rollbacked")
                return False

            except Exception as err:
                transaction.abort()
                logger.exception(u"Unkown Error")
                if mail_addresses:
                    send_unknown_error(
                        self.request, mail_addresses, filename, err
                    )
                    logger.error(
                        u"An error mail has been sent to {0}".format(
                            mail_addresses
                        )
                    )
                logger.exception(u"* FAILED : transaction has been rollbacked")
                return False

        transaction.commit()
        logger.info(u"Measure computing transaction has been commited")
        logger.info(u"* SUCCEEDED !!!")

        if mail_addresses:
            send_success(
                self.request,
                mail_addresses,
                filename,
                num_operations,
                missed_associations,
            )
            logger.info(
                u"A success email has been sent to {0}".format(
                    mail_addresses
                )
            )



# -*- coding: utf-8 -*-
import logging
from pyramid_mailer.message import (
    Attachment,
)

from endi_base.exception import (
    MailAlreadySent,
    UndeliveredMail,
)
from endi_base.mail import send_mail
from endi_celery.models import (
    store_sent_mail,
    check_if_mail_sent,
)


logger = logging.getLogger(__name__)


SALARYSHEET_MAIL_MESSAGE = u"""Bonjour,
Vous trouverez ci-joint votre bulletin de salaire.
"""

SALARYSHEET_MAIL_SUBJECT = u"Votre bulletin de salaire"


def send_salary_sheet(
    request, email, company_id, filename, filepath, force=False,
    message=None, subject=None
):
    """
    Send a salarysheet to the given company's e-mail

    :param obj request: A pyramid request object
    :param str company_mail: The mail to send it to
    :param int company_id: The id of the associated company
    :param str filepath: The path to the filename
    :param bool force: Whether to force sending this file again
    :param str message: The mail message
    :param str subject: The mail subject
    :returns: A MailHistory instance
    :TypeError UndeliveredMail: When the company has no mail
    :TypeError MailAlreadySent: if the file has
        already been sent and no force option was passed
    """
    filebuf = file(filepath, 'r')
    filedatas = filebuf.read()

    if not force and check_if_mail_sent(filedatas, company_id):
        logger.warn(u"Mail already sent : mail already sent")
        raise MailAlreadySent(u"Mail already sent")

    filebuf.seek(0)

    if email is None:
        logger.warn(
            u"Undelivered email : no mail provided for company {0}".format(
                company_id
            )
        )
        raise UndeliveredMail(u"no mail provided for company {0}".format(
            company_id)
        )
    else:
        logger.info('Sending the file %s' % filepath)
        logger.info("Sending it to %s" % email)
        attachment = Attachment(filename, "application/pdf", filebuf)

        subject = subject or SALARYSHEET_MAIL_SUBJECT
        message = message or SALARYSHEET_MAIL_MESSAGE

        send_mail(
            request,
            email,
            message,
            subject,
            attachment,
        )
        return store_sent_mail(filepath, filedatas, company_id)



# -*- coding: utf-8 -*-
from sqlalchemy import desc, func
from endi_base.utils.renderers import configure_export
from endi.compute.math_utils import integer_to_amount
from endi.models.user.userdatas import UserDatas
from endi.models.third_party.customer import Customer
from endi.models.third_party.supplier import Supplier
from endi.models.tva import Tva
from endi.models.task import Task


def _add_userdatas_custom_headers(writer, query):
    """
    Specific to userdatas exports

    Add custom headers that are not added through automation

    Add headers for code_compta
    """
    from endi_base.models.base import DBSESSION
    from endi.models.user.user import COMPANY_EMPLOYEE
    # Compte analytique
    query = DBSESSION().query(
        func.count(COMPANY_EMPLOYEE.c.company_id).label('nb')
    )
    query = query.group_by(COMPANY_EMPLOYEE.c.account_id)
    code_compta_count = query.order_by(desc("nb")).first()
    if code_compta_count:
        code_compta_count = code_compta_count[0]
        for index in range(0, code_compta_count):
            new_header = {
                'label': "Compte_analytique {0}".format(index + 1),
                'name': "code_compta_{0}".format(index + 1),
            }
            writer.add_extra_header(new_header)

    return writer


def _add_userdatas_code_compta(writer, userdatas):
    """
    Add code compta to exports (specific for userdatas exports)

    :param obj writer: The tabbed file writer
    :param obj userdatas: The UserDatas instance we manage
    """
    user_account = userdatas.user
    if user_account:
        datas = []
        for company in user_account.companies:
            datas.append(company.code_compta)
        writer.add_extra_datas(datas)
    return writer


def _add_invoice_custom_headers(writer, invoices):
    """
    Invoice export headers

    * Entreprise ;
    • Client ;
    • Date ;
    • Numéro de factures ;
    • Objet ;
    • Lieu des travaux ;
    • HT (par taux de tva) ;
    • TVA (par taux de tva) ;
    • TTC.
    """
    for tva in Tva.query():
        writer.add_extra_header({
            'label': u"HT {tva.name}".format(tva=tva),
            "name": u"HT {tva.value}".format(tva=tva),
        })
        writer.add_extra_header({'label': tva.name, "name": tva.value})

    writer.add_extra_header({'label': u"Montat TTC", "name": "ttc"})
    return writer


def _add_invoice_datas(writer, invoice):
    ht_values = invoice.tva_ht_parts()
    tva_values = invoice.get_tvas()
    datas = []
    for tva in Tva.query():
        datas.append(
            integer_to_amount(ht_values.get(tva.value, 0), precision=5)
        )
        datas.append(
            integer_to_amount(tva_values.get(tva.value, 0), precision=5)
        )
    datas.append(integer_to_amount(invoice.ttc, precision=5))
    writer.add_extra_datas(datas)
    return writer


def _import_userdatas_add_related_user(action, dbsession, model, updated, args):
    """
    Creat a User instance when importing a new UserDatas instance

    :param str action: An import action (override, insert, update, only_update,
    only_override)
    :param obj dbsession: The database transaction session (DBSESSION)
    :param obj model: The newly inserted model
    :param bool updated: Is it an update of an existing model ?
    :param dict args: The importation arguments
    """
    if action == 'insert':
        model.gen_related_user_instance()
    return model


def _import_third_party_set_label(action, dbsession, model, updated, args):
    """
    Ensure the currently managed model has a label set
    """
    if not model.label:
        model.label = model._get_label()
        model.name = model.label
        model = dbsession.merge(model)
    return model


def includeme(config):
    configure_export()
    config.register_import_model(
        key='userdatas',
        model=UserDatas,
        label=u"Données de gestion sociale",
        permission='admin_userdatas',
        excludes=(
            'name',
            'created_at',
            'updated_at',
            'type_',
            '_acl',
            'parent_id',
            'parent',
        ),
        callbacks=[_import_userdatas_add_related_user]
    )
    config.register_import_model(
        key='customers',
        model=Customer,
        label=u"Clients",
        permission='add_customer',
        excludes=(
            'created_at',
            'updated_at',
            'company_id',
            'company',
        ),
        callbacks=[_import_third_party_set_label]
    )
    config.register_import_model(
        key='suppliers',
        model=Supplier,
        label=u"Fournisseurs",
        permission='add_supplier',
        excludes=(
            'created_at',
            'updated_at',
            'company_id',
            'company',
        ),
        callbacks=[_import_third_party_set_label]
    )
    config.register_export_model(
        key='userdatas',
        model=UserDatas,
        options={
            'hook_init': _add_userdatas_custom_headers,
            'hook_add_row': _add_userdatas_code_compta,
            'foreign_key_name': 'userdatas_id',
        }
    )
    config.register_export_model(
        key='invoices',
        model=Task,
        options={
            'hook_init': _add_invoice_custom_headers,
            'hook_add_row': _add_invoice_datas,
            'foreign_key_name': 'task_id',
            'excludes': ('name', 'created_at', 'updated_at', 'type_'),
            'order': (
                'company',
                'customer',
                'date',
                'official_number',
                'description',
                'workplace',
            )
        }
    )



# -*- coding: utf-8 -*-
"""
Celery tasks used to asynchronously generate exports (like excel exports)


Workflow :
    user provide filters
    TODO : user provide columns

    For UserDatas exports, we need to add some fields


    1- Task entry
    2- retrieve model
    3- generate the file or re-use the cached one
"""
import os
import transaction
from tempfile import mktemp

from pyramid_celery import celery_app
from beaker.cache import cache_region
from celery.utils.log import get_task_logger

from sqla_inspect.ods import SqlaOdsExporter
from sqla_inspect.excel import SqlaXlsExporter
from sqla_inspect.csv import SqlaCsvExporter
from sqlalchemy.orm import (
    RelationshipProperty,
)
from sqlalchemy.sql.expression import label
from sqlalchemy import (
    desc,
    func,
)
from endi_celery.conf import get_setting
from endi_celery.models import FileGenerationJob
from endi_celery.tasks import utils

MODELS_CONFIGURATION = {}

logger = utils.get_logger(__name__)


GENERATION_ERROR_MESSAGE = (
    u"Une erreur inconnue a été rencontrée à la génération de votre fichier, "
    "veuillez contacter votre administrateur en lui"
    "fournissant l'identifiant suivant : %s"
)


def _add_o2m_headers_to_writer(writer, query, id_key):
    """
    Add column headers in the form "label 1",  "label 2" ... to be able to
    insert the o2m related elements to a main model's table export (allow to
    have 3 dimensionnal datas in a 2d array)

    E.g : Userdatas objects have got a o2m relationship on DateDatas objects

    Here we would add date 1, date 2... columns regarding the max number of
    configured datas (if a userdatas has 5 dates, we will have 5 columns)
    We fill the column with the value of an attribute of the DateDatas model
    (that is handled by sqla_inspect thanks to the couple index + related_key
    configuration)

    The name of the attribute is configured using the "flatten" key in the
    relationship's export configuration

    :param str id_key: The foreign key attribute mostly matching the class we
    export (e.g : when exporting UserDatas, most of the related elements point
    to it through a userdatas_id foreign key)
    """
    from endi_base.models.base import DBSESSION
    new_headers = []
    for header in writer.headers:
        if isinstance(header['__col__'], RelationshipProperty):
            if header['__col__'].uselist:
                class_ = header['__col__'].mapper.class_
                # On compte le nombre maximum d'objet lié que l'on rencontre
                # dans la base
                if not hasattr(class_, id_key):
                    continue

                count = DBSESSION().query(
                    label("nb", func.count(class_.id))
                ).group_by(getattr(class_, id_key)).order_by(
                    desc("nb")).first()

                if count is not None:
                    count = count[0]
                else:
                    count = 0

                # Pour les relations O2M qui ont un attribut flatten de
                # configuré, On rajoute des colonnes "date 1" "date 2" dans
                # notre sheet principale
                for index in range(0, count):
                    if 'flatten' in header:
                        flatten_keys = header['flatten']
                        if not hasattr(flatten_keys, '__iter__'):
                            flatten_keys = [flatten_keys]

                        for flatten_key, flatten_label in flatten_keys:
                            new_header = {
                                '__col__': header['__col__'],
                                'label': u"%s %s %s" % (
                                    header['label'],
                                    flatten_label,
                                    index + 1),
                                'key': header['key'],
                                'name': u"%s_%s_%s" % (
                                    header['name'],
                                    flatten_key,
                                    index + 1
                                ),
                                'related_key': flatten_key,
                                'index': index
                            }
                            new_headers.append(new_header)

    writer.headers.extend(new_headers)
    return writer


def _get_tmp_directory_path():
    """
    Return the tmp filepath configured in the current configuration
    :param obj request: The pyramid request object
    """
    asset_path_spec = get_setting('endi.static_tmp', mandatory=True)
    return asset_path_spec


def _get_tmp_filepath(directory, basename, extension):
    """
    Return a temp filepath for the given filename

    :param str basename: The base name to use
    :returns: A path to a non existing file
    :rtype: str
    """
    if not extension.startswith('.'):
        extension = u'.' + extension

    filepath = mktemp(prefix=basename, suffix=extension, dir=directory)
    while os.path.exists(filepath):
        filepath = mktemp(prefix=basename, suffix=extension, dir=directory)
    return filepath


@cache_region('default_term')
def _write_file_on_disk(tmpdir, model_type, ids, filename, extension):
    """
    Return a path to a generated file

    :param str tmpdir: The path to write to
    :param str model_type: The model key we want to generate an ods file for
    :param list ids: An iterable containing all ids of models to be included in
    the output
    :param str filename: The path to the file output
    :param str extension: The desired extension (xls/ods)
    :returns: The name of the generated file (unique and temporary name)
    :rtype: str
    """
    logger.debug(" No file was cached yet")
    config = MODELS_CONFIGURATION[model_type]
    model = config['factory']
    query = model.query()
    if ids:
        query = query.filter(model.id.in_(ids))

    options = {}
    if 'excludes' in config:
        options['excludes'] = config['excludes']
    if 'order' in config:
        options['order'] = config['order']
    if extension == 'ods':
        writer = SqlaOdsExporter(model=model, **options)
    elif extension == 'xls':
        writer = SqlaXlsExporter(model=model, **options)
    elif extension == 'csv':
        writer = SqlaCsvExporter(model=model, **options)

    writer = _add_o2m_headers_to_writer(
        writer,
        query,
        config['foreign_key_name']
    )

    if 'hook_init' in config:
        writer = config['hook_init'](writer, query)

    for item in query:
        writer.add_row(item)
        if 'hook_add_row' in config:
            config['hook_add_row'](writer, item)

    filepath = _get_tmp_filepath(tmpdir, filename, extension)
    logger.debug(" + Writing file to %s" % filepath)

    with open(filepath, 'w') as f_buf:
        writer.render(f_buf)
    return os.path.basename(filepath)


@celery_app.task(bind=True)
def export_to_file(self, job_id, model_type, ids, filename='test',
                   file_format='ods'):
    """
    Export the datas provided in the given query to ods format and generate a

    :param int job_id: The id of the job object used to record file_generation
    informations
    :param str model_type: The model we want to export (see MODELS)
    :param list ids: List of ids to query
    :param str filename: The base filename to use for the export (unique string
    is appended)
    :param str file_format: The format in which we want to export
    """
    logger = get_task_logger(__name__)
    logger.info(u"Exporting to a file")
    logger.info(u" + model_type : %s", model_type)
    logger.info(u" + ids : %s", ids)

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    # Execute actions
    try:
        tmpdir = _get_tmp_directory_path()
        result_filename = _write_file_on_disk(
            tmpdir,
            model_type,
            ids,
            filename,
            file_format,
        )
        logger.debug(u" -> The file %s been written", result_filename)
        transaction.commit()
    except:
        transaction.abort()
        logger.exception("Error while generating file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(
            FileGenerationJob, job_id, errors
        )
    else:
        utils.record_completed(
            FileGenerationJob, job_id, filename=result_filename
        )

    return ""



# -*- coding: utf-8 -*-
"""
All asynchronous tasks runned through enDI are stored here
Tasks are handled by a celery service
Redis is used as the central bus
"""
import time
import transaction
from pyramid.threadlocal import get_current_request

from pyramid_celery import celery_app

from endi_base.exception import (
    UndeliveredMail,
    MailAlreadySent,
)
from endi_celery.mail import send_salary_sheet

from endi_celery.tasks import utils
from endi_celery.models import (
    MailingJob,
)


logger = utils.get_logger(__name__)


def _mail_format_message(mail_message_tmpl, company, kwds):
    """
    Return the message to be sent to a single company
    :param str mail_message_tmpl: Template for the mail message
    :param obj company: The company object
    :param dict kwds: Additionnal keywords to pass to the string.format method
    """
    kwds['company'] = company
    message = mail_message_tmpl.format(**kwds)
    return message


@celery_app.task(bind=True)
def async_mail_salarysheets(self, job_id, mails, force):
    """
    Asynchronously sent a bunch of emails with attached salarysheets

    :param int job_id: The id of the MailSendJob
    :param mails: a list of dict compound of
        {
            'id': company_id,
            'attachment': attachment filename,
            'attachment_path': attachment filepath,
            'message': The mail message,
            'subject': The mail subject,
            'company_id': The id of the company,
            'email': The email to send it to,
        }
    :param force: Should we force the mail sending
    """
    logger.info(u"We are launching an asynchronous mail sending operation")
    logger.info(u"  The job id : %s" % job_id)

    from endi_base.models.base import DBSESSION
    # Mark job started
    utils.start_job(self.request, MailingJob, job_id)

    mail_count = 0
    error_count = 0
    error_messages = []
    request = get_current_request()
    session = DBSESSION()
    for mail_datas in mails:
        transaction.begin()
        # since we send a mail out of the transaction process, we need to commit
        # each mail_history instance to avoid sending and not storing the
        # history
        try:
            company_id = mail_datas['company_id']
            email = mail_datas['email']

            if email is None:
                logger.error(u"no mail found for company {0}".format(
                    company_id)
                )
                continue
            else:
                message = mail_datas['message']
                subject = mail_datas['subject']
                logger.info(u"  The mail subject : %s" % subject)
                logger.info(u"  The mail message : %s" % message)

                mail_history = send_salary_sheet(
                    request,
                    email,
                    company_id,
                    mail_datas['attachment'],
                    mail_datas['attachment_path'],
                    force=force,
                    message=message,
                    subject=subject,
                )
                # Stores the history of this sent email
                session.add(mail_history)
            transaction.commit()

        except MailAlreadySent as e:
            transaction.abort()
            error_count += 1
            msg = u"Ce fichier a déjà été envoyé {0}".format(
                mail_datas['attachment']
            )
            error_messages.append(msg)
            logger.exception(u"Mail already delivered")
            logger.error(u"* Part of the Task FAILED")
            continue

        except UndeliveredMail as e:
            transaction.abort()
            error_count += 1
            msg = u"Impossible de délivrer de mail à l'entreprise {0} \
(mail : {1})".format(company_id, email)
            error_messages.append(msg)
            logger.exception(u"Unable to deliver an e-mail")
            logger.error(u"* Part of the Task FAILED")
            continue

        except Exception as e:
            transaction.abort()
            error_count += 1
            logger.exception(u"The subtransaction has been aborted")
            logger.error(u"* Part of the task FAILED !!!")
            error_messages.append(u"{0}".format(e))

        else:
            mail_count += 1
            logger.info(u"The transaction has been commited")
            logger.info(u"* Part of the Task SUCCEEDED !!!")

    messages = [u"{0} mails ont été envoyés".format(mail_count)]
    messages.append(
        u"{0} mails n'ont pas pu être envoyés".format(error_count)
    )

    logger.info(u"-> Task finished")
    if error_count > 0:
        utils.record_failure(
            MailingJob,
            job_id,
            error_messages=error_messages,
            messages=messages,
        )
    else:
        utils.record_completed(
            MailingJob,
            job_id,
            messages=messages,
        )



# -*- coding: utf-8 -*-
"""
Tasks used to compile treasury measures
"""
import transaction

from pyramid_celery import celery_app
from sqlalchemy import or_, desc
from collections import OrderedDict

from endi_base.models.base import DBSESSION
from endi_base.mail import send_mail
from endi.models.accounting.operations import (
    AccountingOperationUpload,
    AccountingOperation,
)
from endi.models.accounting.treasury_measures import (
    TreasuryMeasure,
    TreasuryMeasureGrid,
    TreasuryMeasureType,
    TreasuryMeasureTypeCategory,
)
from endi_celery.conf import (
    get_recipients_addresses,
)
from endi.models.accounting.income_statement_measures import (
    IncomeStatementMeasure,
    IncomeStatementMeasureGrid,
    IncomeStatementMeasureType,
    IncomeStatementMeasureTypeCategory,
)

from endi_celery.tasks import utils


logger = utils.get_logger(__name__)


class GridCacheItem:
    """
    Wraps a grid in cache and provide storages for

    measures
    total value by measure_type
    total value by measure type category
    """
    def __init__(self, grid):
        self.grid = grid
        self._category_totals = {}
        self._type_totals = {}
        self._measures = {}

    def clean_old_measures(self, type_ids):
        """
        Clean old measures

        Set them to  or remove them if theire type is not in the type list

        :param list type_ids: List of ids (int)
        :returns: The list of measures to remove
        """
        old_measures = [
            measure for measure in self.grid.measures
            if measure.measure_type_id not in type_ids
        ]
        self.grid.measures = [
            measure for measure in self.grid.measures
            if measure.measure_type_id in type_ids
        ]

        for measure in self.grid.measures:
            measure.value = 0
            self._measures[measure.measure_type_id] = measure
        return old_measures

    def get_measures(self):
        """
        Collect all existing measures (newly created or already existing)
        """
        return self._measures

    def store_measure(self, measure):
        """
        Store a measure in the measure cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        self._measures[measure.measure_type_id] = measure
        if measure not in self.grid.measures:
            self.grid.measures.append(measure)

    def set_total_type(self, measure):
        """
        Store the total value for a given measure type in the grid's cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        self._type_totals[measure.measure_type_id] = measure.value

    def update_category_total(self, measure):
        """
        Update the category total value for a given measure type in the grid's
        cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        measure_type = measure.measure_type
        if not measure_type.computed_total:
            if measure_type.category_id not in self._category_totals:
                self._category_totals[measure_type.category_id] = 0
            self._category_totals[measure_type.category_id] += measure.value

    def get_type_total(self, typ_id):
        """
        Get the total type for a given measure type

        :param int typ_id: The id of the measure type
        """
        return self._type_totals.get(typ_id, 0)

    def get_category_total(self, category_id):
        """
        Get the category total for a given measure type category

        :param int category_id: The id of the category
        """
        return self._category_totals.get(category_id, 0)


class GridCache:
    """
    A grid cache used to temporarly store all datas related to the current
    computation

    Eache item is an Accounting Measure Grid wraped with the GridCacheItem
    object
    """
    def __init__(self):
        self._datas = {}

    def store_grid(self, key, grid):
        self._datas[key] = GridCacheItem(grid)
        return self._datas[key]

    def get(self, key):
        return self._datas.get(key)


class BaseMeasureCompiler(object):
    """
    Base measure compiler

    Uses a cache to store grids currently edited

    For each grid, the cache contains :

        - Measures already created
        - totals by category
        - totals by type

    Totals are used in a second time to compute "total" measures (not related to
    account numbers)
    """
    measure_type_class = None
    measure_type_category_class = None
    measure_grid_class = None
    measure_class = None

    def __init__(self, upload, operations):
        self.upload = upload
        self.filetype = upload.filetype
        self.operations = operations
        self.session = DBSESSION()
        self.cache = GridCache()

        # Collect all types
        self.common_measure_types = self._collect_common_measure_types()
        self.categories = self._collect_categories()
        self.computed_measure_types = self._collect_computed_measure_types()

        self.all_type_ids = [t.id for t in self.common_measure_types]
        for cat_id, types in self.computed_measure_types.items():
            self.all_type_ids.extend([t.id for t in types])

        self.processed_grids = []

    def _get_active_type_query(self):
        """
        Build a query to list active measure types
        :returns: An sqlalchemy query object
        """
        return self.measure_type_class.query(
        ).join(
            self.measure_type_category_class
        ).filter(
            self.measure_type_class.active == True  # noqa: E712
        ).filter(
            self.measure_type_category_class.active == True  # noqa: E712
        )

    def _collect_common_measure_types(self):
        """
        Retrieve measure types that are not computed from other measures
        :returns: The list of measure types
        """
        return self._get_active_type_query().filter(
            or_(
                self.measure_type_class.is_total == False,  # noqa: E712
                self.measure_type_class.total_type == 'account_prefix',
            )
        ).all()

    def _collect_categories(self):
        """
        Retrieve all measure type categories
        """
        return self.measure_type_category_class.get_categories()

    def _collect_computed_measure_types(self):
        """
        Collect computed measure types and store them by category id

        :returns: a dict {category_id: [list of MeasureType]}
        """
        result = OrderedDict()
        query = self._get_active_type_query()
        query = query.filter(
            self.measure_type_class.is_total == True,  # noqa: E712
        ).filter(
            self.measure_type_class.total_type != 'account_prefix',
        ).order_by(
            self.measure_type_category_class.order
        ).order_by(self.measure_type_class.order)
        for typ in query:
            result.setdefault(typ.category_id, []).append(typ)
        return result

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object

        :param obj grid: The grid instance
        :returns: A valid dict key
        """
        pass

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object

        :param obj operation: The AccountingOperation instance
        :returns: A valid dict key
        """
        pass

    def _get_new_measure(self, measure_type, grid_id):
        """
        Build a new measure
        """
        measure = self.measure_class(
            label=measure_type.label,
            grid_id=grid_id,
            measure_type_id=measure_type.id,
            order=measure_type.order,
        )
        self.session.add(measure)
        self.session.flush()
        return measure

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load
        specific to the Measure Types

        :param obj grid_item: A GridCacheItem instance
        """
        pass

    def get_grid_item(self, operation):
        """
        Retrieve the grid related to the given operation datas

        :param obj operation: an AccountingOperation instance
        :returns: A Grid instance
        """
        key = self.get_cache_key_from_operation(operation)
        grid_item = self.cache.get(key)
        if grid_item is None:
            grid = self._query_grid(operation)
            if grid is None:
                grid = self._get_new_grid(operation)
                self.session.add(grid)
                self.session.flush()
                grid_item = self.cache.store_grid(key, grid)
            else:
                grid_item = self.cache.store_grid(key, grid)
                self._clean_existing_grid(grid_item)
        return grid_item

    def get_values_for_computation(self, grid_item):
        """
        Collect total values in a dict for complex datas computation

        :param obj grid_item: An instance of GridCacheItem
        """
        result = {}
        for category in self.categories:
            result[category.label] = grid_item.get_category_total(category.id)

        for typ_ in self.common_measure_types:
            result[typ_.label] = grid_item.get_type_total(typ_.id)

        for typlist_ in self.computed_measure_types.values():
            for typ_ in typlist_:
                value = grid_item.get_type_total(typ_.id)
                result[typ_.label] = value

        return result

    def _cache_grid_totals(self):
        """
        Cache all grid totals in order to use them in computation
        """
        for grid_item in self.processed_grids:
            for measure in grid_item.get_measures().values():
                grid_item.set_total_type(measure)
                grid_item.update_category_total(measure)

    def _process_common_measures(self):
        """
        Compile common measures (related to an account prefix) with the given
        operations

        :returns: A list of GridCacheItem
        """
        logger.debug(
            u"    + Processing datas with {}".format(self.__class__.__name__)
        )
        for operation in self.operations:
            if operation.company_id is None:
                continue

            current_grid_item = self.get_grid_item(operation)
            grid = current_grid_item.grid
            if current_grid_item not in self.processed_grids:
                self.processed_grids.append(current_grid_item)

            measures = current_grid_item.get_measures()

            for measure_type in self.common_measure_types:
                matched = False
                if measure_type.match(operation.general_account):
                    measure = measures.get(measure_type.id)
                    if measure is None:
                        measure = self._get_new_measure(
                            measure_type,
                            grid.id,
                        )
                        current_grid_item.store_measure(measure)

                    measure.value += operation.total()
                    matched = True

                if matched:
                    self.session.merge(measure)
                    self.session.flush()

        return self.processed_grids

    def _complete_grids_with_common_measures(self):
        """
        Insert common measures in the grids if not set yet
        """
        for grid_item in self.processed_grids:
            measures = grid_item.get_measures()
            for measure_type in self.common_measure_types:
                if measure_type.id not in measures:
                    measure = self._get_new_measure(
                        measure_type,
                        grid_item.grid.id,
                    )
                    grid_item.store_measure(measure)

    def _process_computed_measures(self):
        """
        Process dynamically computed measures (based on the other ones)

        :returns: A list of GridCacheItem
        """
        for grid_item in self.processed_grids:
            measures = grid_item.get_measures()
            for category_id, measure_types in \
                    self.computed_measure_types.items():
                for measure_type in measure_types:
                    values = self.get_values_for_computation(grid_item)
                    value = measure_type.compute_total(values)
                    measure = measures.get(measure_type.id)
                    if measure is None:
                        measure = self._get_new_measure(
                            measure_type,
                            grid_item.grid.id,
                        )
                        grid_item.store_measure(measure)

                    measure.value = value
                    self.session.merge(measure)
                    grid_item.set_total_type(measure)

        return self.processed_grids

    def process_datas(self):
        """
        Main entry point to process measure computation
        """
        self._process_common_measures()
        self._complete_grids_with_common_measures()
        self.session.flush()
        self._cache_grid_totals()
        self._process_computed_measures()
        self.session.flush()
        return [griditem.grid for griditem in self.processed_grids]


class TreasuryMeasureCompiler(BaseMeasureCompiler):
    measure_type_class = TreasuryMeasureType
    measure_type_category_class = TreasuryMeasureTypeCategory
    measure_grid_class = TreasuryMeasureGrid
    measure_class = TreasuryMeasure
    label = u"Génération des états de trésorerie"

    def get_message(self, grids):
        return u"""Génération des états de trésorerie

États de trésorerie générés : {}

----
enDI
""".format(len(grids))

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object
        """
        key = (operation.company_id, self.upload.updated_at.date())
        return key

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object
        """
        key = (grid.company_id, grid.date)
        return key

    def _query_grid(self, operation):
        """
        Query a grid associated to the given operation

        Query filters should match the get_cache_key_from_operation options
        """
        query = TreasuryMeasureGrid.query().filter_by(
            date=self.upload.updated_at.date()
        ).filter_by(company_id=operation.company_id)

        query = query.order_by(
            desc(TreasuryMeasureGrid.datetime)
        )
        return query.first()

    def _get_new_grid(self, operation):
        """
        Build a new grid based on the given operation

        :param obj operation: The AccountingOperation from which we build
        measure
        """
        return TreasuryMeasureGrid(
            date=self.upload.updated_at.date(),
            company_id=operation.company_id,
            upload=self.upload,
        )

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load

        :param obj grid_item: A GridCacheItem instance
        """
        for measure in grid_item.grid.measures:
            self.session.delete(measure)


class IncomeStatementMeasureCompiler(BaseMeasureCompiler):
    measure_type_class = IncomeStatementMeasureType
    measure_type_category_class = IncomeStatementMeasureTypeCategory
    measure_grid_class = IncomeStatementMeasureGrid
    measure_class = IncomeStatementMeasure
    label = u"Génération des comptes de résultat"

    def get_message(self, grids):
        return u"""Génération des comptes de résultat :

Comptes de résultat traités : {}

----
enDI
""".format(len(grids)/12)

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object
        """
        return (operation.company_id, operation.date.year, operation.date.month)

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object
        """
        return (grid.company_id, grid.year, grid.month)

    def _query_grid(self, operation):
        """
        Query a grid associated to the given operation

        Query filters should match the get_cache_key_from_operation options
        """
        return IncomeStatementMeasureGrid.query().filter_by(
            year=operation.date.year
        ).filter_by(
            month=operation.date.month
        ).filter_by(company_id=operation.company_id).first()

    def _get_new_grid(self, operation):
        """
        Build a new grid based on the given operation

        :param obj operation: The AccountingOperation from which we build
        measure
        """
        return IncomeStatementMeasureGrid(
            year=operation.date.year,
            month=operation.date.month,
            company_id=operation.company_id,
            upload=self.upload,
        )

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load

        :param obj grid_item: A GridCacheItem instance
        """
        # Only clean the old grid item measures (keep existing)
        old_measures = grid_item.clean_old_measures(self.all_type_ids)
        for measure in old_measures:
            self.session.delete(measure)


def get_measure_compilers(data_type, grid_type=None):
    """
    Retrieve the measure compilers to be used with this given type of datas

    :param str data_type: The type of data we build our measures from
    :returns: The measure compiler
    """
    if data_type == AccountingOperationUpload.ANALYTICAL_BALANCE:
        logger.info("  + Handling analytical_balance file")
        return [TreasuryMeasureCompiler]
    elif data_type == AccountingOperationUpload.GENERAL_LEDGER:
        logger.info("  + Handling General Ledger file")
        return [IncomeStatementMeasureCompiler]
    elif data_type == AccountingOperationUpload.SYNCHRONIZED_ACCOUNTING:
        logger.info("  + Handling Synchronized Datas")
        if grid_type == 'treasury':
            return [TreasuryMeasureCompiler]
        elif grid_type == 'income_statement':
            return [IncomeStatementMeasureCompiler]
        else:
            return [TreasuryMeasureCompiler, IncomeStatementMeasureCompiler]

MAIL_ERROR_SUBJECT = u"[Erreur] {}"
MAIL_SUCCESS_SUBJECT = u"[Succès] {}"


def send_success(request, mail_addresses, type_label, message):
    """
    Send a success email

    :param obj request: The current request object
    :param list mail_addresses: List of recipients e-mails
    :param str message: message to send
    """
    if mail_addresses:
        try:
            subject = MAIL_SUCCESS_SUBJECT.format(type_label)
            send_mail(request, mail_addresses, message, subject)
        except:
            logger.exception(u"Error sending email")


def send_error(request, mail_addresses, type_label, message):
    """
    Send an error email

    :param obj request: The current request object
    :param list mail_addresses: List of recipients e-mails
    :param str message: message to send
    """
    if mail_addresses:
        try:
            subject = MAIL_ERROR_SUBJECT.format(type_label)
            send_mail(request, mail_addresses, message, subject)
        except:
            logger.exception(u"Error sending email")


@celery_app.task(bind=True)
def compile_measures_task(self, upload_id, grid_type=None):
    """
    Celery task handling measures compilation
    """
    logger.info(
        u"Launching the compile measure task for upload {0}".format(upload_id)
    )

    if grid_type is not None:
        logger.info(u"Only compiling {} grids".format(grid_type))

    transaction.begin()
    upload = AccountingOperationUpload.get(upload_id)
    if upload is None:
        raise Exception(
            u"No AccountingOperationUpload instance with id {}".format(
                upload_id
            )
        )
    operations = AccountingOperation.query().filter_by(
        upload_id=upload_id
    ).all()

    mail_addresses = get_recipients_addresses(self.request)
    compiler_factories = get_measure_compilers(upload.filetype, grid_type)
    messages = []
    for compiler_factory in compiler_factories:
        try:
            compiler = compiler_factory(upload, operations)
            grids = compiler.process_datas()
            messages.append(compiler.get_message(grids))
        except Exception as err:
            logger.exception(u"Error while generating measures")
            transaction.abort()
            send_error(
                self.request,
                mail_addresses,
                compiler.label,
                err.message,
            )
            return False
        else:
            logger.info(u"{0} measure grids were handled".format(len(grids)))
            send_success(
                self.request,
                mail_addresses,
                compiler.label,
                compiler.get_message(grids),
            )
            logger.info(
                u"A success email has been sent to {0}".format(
                    mail_addresses
                )
            )

    transaction.commit()
    logger.info(u"The transaction has been commited")
    logger.info(u"* Task SUCCEEDED !!!")



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(here, 'requirements.txt')) as f:
    requires = f.read()

with open(os.path.join(here, 'CURRENT_VERSION')) as f:
    current_version = f.read().splitlines()[0].strip()


entry_points = {
    "paste.app_factory": [
        "worker = endi_celery:worker",
        "scheduler = endi_celery:scheduler",
    ]
}

setup(name='endi_celery',
      version=current_version,
      description='endi_celery',
      long_description=README,
      license='GPLv3',
      classifiers=[
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='Coopérer Pour Entreprendre',
      author_email='contact@endi.coop',
      url='https://framagit.org/endi/endi_celery',
      keywords='web wsgi bfg pylons pyramid celery',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      entry_points=entry_points,
      )



# -*- coding:utf-8 -*-
import locale
locale.setlocale(locale.LC_ALL, "fr_FR.UTF-8")
locale.setlocale(locale.LC_TIME, "fr_FR.UTF-8")

import logging
from pyramid.config import Configurator
from pyramid.paster import setup_logging
from sqlalchemy import engine_from_config
from pyramid_beaker import set_cache_regions_from_settings

from endi.models import *
from endi_base.models.initialize import initialize_sql
from endi_celery.tasks.csv_import import (
    MODELS_CONFIGURATION as IMPORT_MODELS_CONFIGURATION,
)
from endi_celery.tasks.export import (
    MODELS_CONFIGURATION as EXPORT_MODELS_CONFIGURATION
)


def register_import_model(
    config, key, model, label, permission, excludes,
    callbacks=()
):
    """
    Register a model for import

    :param obj config: The pyramid configuration object
    :param str key: The key used to identify the model type
    :param class model: The model to be used
    :param str label: A label describing the type of datas
    :param str permission: The permission to associate to this import
    :param tuple excludes: The field of the model we don't want to handle in the
    import
    :param list callbacks: List of callables called after import action
    """
    IMPORT_MODELS_CONFIGURATION[key] = {
        'factory': model,
        'label': label,
        'permission': permission,
        'excludes': excludes,
        'callbacks': callbacks,
    }


def register_export_model(config, key, model, options={}):
    """
    Register a model for export

    :param obj config: The pyramid configuration object
    :param str key: The key used to identify the model type
    :param class model: The model to be used
    """
    EXPORT_MODELS_CONFIGURATION[key] = {'factory': model}
    EXPORT_MODELS_CONFIGURATION[key].update(options)


def includeme(config):
    """
    Includes some celery specific stuff in the main application
    """
    config.add_directive("register_import_model", register_import_model)
    config.add_directive("register_export_model", register_export_model)
    config.include("endi_celery.tasks")


def worker(global_config, **settings):
    """
    Entry point for the pyramid celery stuff
    """
    logging.basicConfig()
    setup_logging(global_config['__file__'])
    logger = logging.getLogger(__name__)
    logger.info("Bootstraping app")
    engine = engine_from_config(settings, 'sqlalchemy.')
    set_cache_regions_from_settings(settings)
    initialize_sql(engine)
    config = Configurator(settings=settings)
    config.include('pyramid_celery')
    includeme(config)

    config.configure_celery(global_config['__file__'])
    config.commit()
    return config.make_wsgi_app()


def scheduler(global_config, **settings):
    logging.basicConfig()
    setup_logging(global_config['__file__'])
    logger = logging.getLogger(__name__)
    logger.info("Bootstraping celery scheduler application")
    config = Configurator(settings=settings)
    config.include('pyramid_celery')
    config.configure_celery(global_config['__file__'])
    config.commit()
    return config.make_wsgi_app()
